@@ -4,13 +4,13 @@ use strict;
 use warnings;
 use POE qw(Wheel::SocketFactory Wheel::ReadWrite Filter::Stackable Filter::Line Filter::IRCD);
 use POE::Component::Server::IRC::Plugin qw( :ALL );
+use POE::Component::Server::IRC::Pipeline;
 use Socket;
 use Carp;
 use Net::Netmask;
-use base qw(POE::Component::Pluggable);
 use vars qw($VERSION);
 
-$VERSION = '1.22';
+$VERSION = '1.01';
 
 sub create {
   my $package = shift;
@@ -25,15 +25,6 @@ sub create {
   $self->{antiflood} = 1 unless defined $self->{antiflood} and $self->{antiflood} eq '0';
   my $options = delete $self->{options};
   my $sslify_options = delete $self->{sslify_options};
-  my $plugin_debug   = delete $self->{plugin_debug};
-
-  $self->_pluggable_init(
-        prefix     => $self->{prefix},
-        reg_prefix => 'PCSI_',
-        types      => { SERVER => 'IRCD' },
-        ($plugin_debug ? (debug => 1) : () ),
-  );
-
 
   $self->{session_id} = POE::Session->create(
 	object_states => [
@@ -109,7 +100,7 @@ sub call {
 }
 
 sub _start {
-  my ($kernel,$self,$sender) = @_[KERNEL,OBJECT,SENDER];
+  my ($kernel,$self) = @_[KERNEL,OBJECT];
 
   $self->{session_id} = $_[SESSION]->ID();
 
@@ -136,10 +127,6 @@ sub _start {
 	$self->{will_do_auth} = 1;
   }
   $self->_load_our_plugins();
-  if ( $kernel != $sender ) {
-    $self->{sessions}->{ $sender->ID }++;
-    $kernel->post( $sender => $self->{prefix} . 'registered' => $self );
-  }
   undef;
 }
 
@@ -156,7 +143,9 @@ sub register {
   $session = $session->ID(); $sender = $sender->ID();
 
   $self->{sessions}->{ $sender }++;
-  $kernel->refcount_increment( $sender => __PACKAGE__ ) if $self->{sessions}->{ $sender } == 1 and $sender ne $session;
+  if ( $self->{sessions}->{ $sender } == 1 and $sender ne $session ) {
+	$kernel->refcount_increment( $sender => __PACKAGE__ );
+  }
   $kernel->post( $sender => $self->{prefix} . 'registered' => $self );
   undef;
 }
@@ -166,7 +155,9 @@ sub unregister {
   $session = $session->ID(); $sender = $sender->ID();
 
   delete $self->{sessions}->{ $sender };
-  $kernel->refcount_decrement( $sender => __PACKAGE__ ) if $sender ne $session;
+  if ( $sender ne $session ) {
+	$kernel->refcount_decrement( $sender => __PACKAGE__ );
+  }
   $kernel->post( $sender => $self->{prefix} . 'unregistered' );
   undef;
 }
@@ -179,17 +170,28 @@ sub shutdown {
 sub _shutdown {
   my ($kernel,$self) = @_[KERNEL,OBJECT];
 
-  $kernel->alias_remove( $_ ) for $kernel->alias_list();
-  $kernel->refcount_decrement( $self->{session_id} => __PACKAGE__ ) unless $self->{alias};
+  if ( $self->{alias} ) {
+	$kernel->alias_remove( $_ ) for $kernel->alias_list();
+  } else {
+	$kernel->refcount_decrement( $self->{session_id} => __PACKAGE__ );
+  }
 
   $self->{terminating} = 1;
+  # Terminate listeners
   delete $self->{listeners};
+  # Terminate any pending connectors
   delete $self->{connectors};
+  #ToDo: Terminate all connections gracefully and send appropriate disconnect messages
+  #      for servers first then for clients.
+  # Dirty hack
   delete $self->{wheels}; # :)
   $kernel->alarm_remove_all();
+  # Unregister all registered sessions.
   $kernel->refcount_decrement( $_ => __PACKAGE__ ) for keys %{ $self->{sessions} };
-  $self->_pluggable_destroy();
+  #ToDo: unload all loaded plugins.
   $self->_unload_our_plugins();
+  
+  #$kernel->call( $self->{ident_client} => 'shutdown' );
   undef;
 }
 
@@ -209,32 +211,28 @@ sub send_event {
 }
 
 sub __send_event {
-  my( $self, $event, @args ) = @_[ OBJECT, ARG0, ARG1 .. $#_ ];
-  $self->_send_event( $event, @args );
-  return 1;
+	my( $self, $event, @args ) = @_[ OBJECT, ARG0, ARG1 .. $#_ ];
+
+	# Actually send the event...
+	$self->_send_event( $event, @args );
+	return 1;
 }
 
 sub _send_event {
   my ($self,$event,@args) = @_;
-  return 1 if $self->_pluggable_process( 'SERVER', $event, \( @args ) ) == PCSI_EAT_ALL;
+  # Let the plugin system process this
+  return 1 if $self->_plugin_process( 'SERVER', $event, \( @args ) ) == PCSI_EAT_ALL;
   $poe_kernel->post( $_ => $event, @args ) for  keys % { $self->{sessions} };
   return 1;
 }
 
-sub _pluggable_event {
-    my ($self, @args) = @_;
-    $self->yield(__send_event => @args);
-    return;
-}
-
 ############################
 # Listener related methods #
 ############################
 
 sub _accept_failed {
-  my ($kernel,$self,$operation,$errnum,$errstr,$listener_id) = @_[KERNEL,OBJECT,ARG0..ARG3];
+  my ($kernel,$self,$listener_id) = @_[KERNEL,OBJECT,ARG3];
   delete $self->{listeners}->{ $listener_id };
-  $self->_send_event( $self->{prefix} . 'listener_failure', $listener_id, $operation, $errnum, $errstr );
   undef;
 }
 
@@ -301,7 +299,7 @@ sub _add_listener {
 
   my $listener = POE::Wheel::SocketFactory->new(
 	BindPort => $bindport,
-	( $parms{bindaddr} ? ( BindAddress => $parms{bindaddr} ) : () ),
+	( $parms{bindaddr} ? ( BindAddr => $parms{bindaddr} ) : () ),
 	Reuse => 'on',
 	( $parms{listenqueue} ? ( ListenQueue => $parms{listenqueue} ) : () ),
 	SuccessEvent => '_accept_connection',
@@ -905,6 +903,188 @@ sub exempted {
 # Plugin methods #
 ##################
 
+# accesses the plugin pipeline
+sub pipeline {
+  my ($self) = @_;
+  $self->{PLUGINS} = POE::Component::Server::IRC::Pipeline->new($self)
+    unless UNIVERSAL::isa($self->{PLUGINS}, 'POE::Component::Server::IRC::Pipeline');
+  return $self->{PLUGINS};
+}
+
+# Adds a new plugin object
+sub plugin_add {
+  my ($self, $name, $plugin) = @_;
+  my $pipeline = $self->pipeline;
+
+  unless (defined $name and defined $plugin) {
+    warn 'Please supply a name and the plugin object to be added!';
+    return;
+  }
+
+  return $pipeline->push($name => $plugin);
+}
+
+# Removes a plugin object
+sub plugin_del {
+  my ($self, $name) = @_;
+
+  unless (defined $name) {
+    warn 'Please supply a name/object for the plugin to be removed!';
+    return;
+  }
+
+  my $return = scalar $self->pipeline->remove($name);
+  warn "$@\n" if $@;
+  return $return;
+}
+
+# Gets the plugin object
+sub plugin_get {
+  my ($self, $name) = @_;  
+
+  unless (defined $name) {
+    warn 'Please supply a name/object for the plugin to be removed!';
+    return;
+  }
+
+  return scalar $self->pipeline->get($name);
+}
+
+# Lists loaded plugins
+sub plugin_list {
+  my ($self) = @_;
+  my $pipeline = $self->pipeline;
+  my %return;
+
+  for (@{ $pipeline->{PIPELINE} }) {
+    $return{ $pipeline->{PLUGS}{$_} } = $_;
+  }
+
+  return \%return;
+}
+
+# Lists loaded plugins in order!
+sub plugin_order {
+  my ($self) = @_;
+  return $self->pipeline->{PIPELINE};
+}
+
+# Lets a plugin register for certain events
+sub plugin_register {
+  my ($self, $plugin, $type, @events) = @_;
+  my $pipeline = $self->pipeline;
+
+  unless (defined $type and ($type eq 'SERVER' or $type eq 'USER')) {
+    warn 'Type should be SERVER or USER!';
+    return;
+  }
+
+  unless (defined $plugin) {
+    warn 'Please supply the plugin object to register!';
+    return;
+  }
+
+  unless (@events) {
+    warn 'Please supply at least one event to register!';
+    return;
+  }
+
+  for my $ev (@events) {
+    if (ref($ev) and ref($ev) eq "ARRAY") {
+      @{ $pipeline->{HANDLES}{$plugin}{$type} }{ map lc, @$ev } = (1) x @$ev;
+    }
+    else {
+      $pipeline->{HANDLES}{$plugin}{$type}{lc $ev} = 1;
+    }
+  }
+
+  return 1;
+}
+
+# Lets a plugin unregister events
+sub plugin_unregister {
+  my ($self, $plugin, $type, @events) = @_;
+  my $pipeline = $self->pipeline;
+
+  unless (defined $type and ($type eq 'SERVER' or $type eq 'USER')) {
+    warn 'Type should be SERVER or USER!';
+    return;
+  }
+
+  unless (defined $plugin) {
+    warn 'Please supply the plugin object to register!';
+    return;
+  }
+
+  unless (@events) {
+    warn 'Please supply at least one event to unregister!';
+    return;
+  }
+
+  for my $ev (@events) {
+    if (ref($ev) and ref($ev) eq "ARRAY") {
+      for my $e (map lc, @$ev) {
+        unless (delete $pipeline->{HANDLES}{$plugin}{$type}{$e}) {
+          warn "The event '$e' does not exist!";
+          next;
+        }
+      }
+    }
+    else {
+      $ev = lc $ev;
+      unless (delete $pipeline->{HANDLES}{$plugin}{$type}{$ev}) {
+        warn "The event '$ev' does not exist!";
+        next;
+      }
+    }
+  }
+
+  return 1;
+}
+
+# Process an input event for plugins
+sub _plugin_process {
+  my ($self, $type, $event, @args) = @_;
+  my $pipeline = $self->pipeline;
+
+  $event = lc $event;
+  my $prefix = $self->{prefix};
+  $event =~ s/^\Q$prefix\E//;
+
+  my $sub = ($type eq 'SERVER' ? "IRCD" : "U") . "_$event";
+  my $return = PCSI_EAT_NONE;
+  my $our_ret = $return;
+
+  if ( $self->can($sub) ) {
+  	$our_ret = $self->$sub( $self, @args );
+  } elsif ( $self->can('_default') ) {
+	$our_ret = $self->_default( $self, $sub, @args );
+  }
+
+  return $return if $our_ret == PCSI_EAT_PLUGIN;
+  $return = PCSI_EAT_ALL if $our_ret == PCSI_EAT_CLIENT;
+  return PCSI_EAT_ALL if $our_ret == PCSI_EAT_ALL;
+
+  for my $plugin (@{ $pipeline->{PIPELINE} }) {
+    next
+      unless $pipeline->{HANDLES}{$plugin}{$type}{$event}
+      or $pipeline->{HANDLES}{$plugin}{$type}{all};
+
+    my $ret = PCSI_EAT_NONE;
+
+    eval { $ret = $plugin->$sub($self, @args) };
+    warn "$sub call failed with $@\n" if $@ and $self->{plugin_debug} and $plugin->can($sub);
+    eval { $ret = $plugin->_default($self, $sub, @args) } if $@;
+    warn "_default call failed with $@\n" if $@ and $self->{plugin_debug};
+
+    return $return if $ret == PCSI_EAT_PLUGIN;
+    $return = PCSI_EAT_ALL if $ret == PCSI_EAT_CLIENT;
+    return PCSI_EAT_ALL if $ret == PCSI_EAT_ALL;
+  }
+
+  return $return;
+}  
+
 1;
 __END__
 
@@ -957,9 +1137,6 @@ Returns an object. Accepts the following parameters, all are optional:
 	antiflood => 0, # Disable flood protection globally, default enabled.
   );
 
-If the component is created from within another session, that session will be automagcially registered
-with the component to receive events and get an 'ircd_backend_registered' event.
-
 =back
 
 =head1 METHODS
@@ -1139,16 +1316,6 @@ Once registered your session will receive these states, which will have the appl
 	ARG0, the listening port;
 	ARG1, the listener id;
 
-=item listener_failure 
-
-  Emitted: when a listener wheel fails;
-  Target: all plugins and registered sessions;
-  Args:
-	ARG0, the listener id;
-	ARG1, the name of the operation that failed;
-	ARG2, numeric value for $!;
-	ARG3, string value for $!;
-
 =item socketerr
 
   Emitted: on the failure of an add_connector()
@@ -1285,12 +1452,6 @@ And plugin related states, prefixed with the appropriate prefix or the default,
 
 Chris 'BinGOs' Williams
 
-=head1 LICENSE
-
-Copyright C<(c)> Chris Williams
-
-This module may be used, modified, and distributed under the same terms as Perl itself. Please see the license that came with your Perl distribution for details.
-
 =head1 SEE ALSO
 
 L<POE::Filter::IRCD>
